=begin

Permission to use, copy, modify, and distribute this software for
any purpose and without fee is hereby granted, provided that the above
copyright notice appear in all copies.

THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.

Name:         ae_UnityPreviews.rb
Author:       Andreas Eisenbarth
Usage:        Should execute automatically on new models or when a model is saved.

Description:  This plugin creates large thumbnails for Ubuntu Unity.

              AE::UnityPreviews.create(model)
                Creates and saves a preview and thumbnail image for the given model.
                The file name of the thumbnail is the md5 sum of the uri of the file
                it is referring to (Free Desktop Standard).
                
              AE::UnityPreviews.get_thumb_path(model_path, &block)
                This method gets the file path for the thumbnail. Since SketchUp does
                not do any effort to support Linux natively, we need some hacky tricks
                to break out of Wine and get a native Unix uri instead of a Windows path.

              Usually thumbnailer programs would be registered in Gnome and then
              envoked by the system when needed.
              [Gnome 2] http://developer.gnome.org/integration-guide/stable/thumbnailer.html.en

              However we can only create large images from SketchUp files when SketchUp 
              is running, and launching SketchUp takes so much time that we don't want
              to that from the thumbnailer program.
              So we generate thumbnails for every file that is opened in SketchUp
              and store the thumbnail in the thumbnail cache.
              According to the Free Desktop Specification the thumbnail cache is in
              $XDG_CACHE_HOME/thumbnails/normal/<md5 hash of target file URI>.png
              $XDG_CACHE_HOME/thumbnails/large/<md5 hash of target file URI>.png
              (~/.thumbnails or ~/.cache/thumbnails)
              http://specifications.freedesktop.org/thumbnail-spec/thumbnail-spec-latest.html#THUMBSAVE

              As it seems, Ubuntu Unity does not make use of these large thumbnails, but
              keeps only a session cache in ~/.local/share/previews using a shorter
              C++ std::hash<std::string> hash_fn. Since we can not create skp
              previews on demand without launching SketchUp and since we cannot
              predict or reproduce the C++ hash, we register a Gnome thumbnailer
              that copies the image file from the Free Desktop thumbnail cache
              to the Unity preview cache.

Date:         24.10.2012
Version:      0.2.1
=end
require 'sketchup.rb'



module AE



module UnityPreviews

  temp = [ENV['TMPDIR'], ENV['TMP'], ENV['TEMP'], ENV['USERPROFILE'], '.'].inject(nil){ |t,dir| (!t && dir && File.directory?(dir) && File.writable?(dir))? dir : t }
  @temp = File.expand_path(temp)

  OSX = ( Object::RUBY_PLATFORM =~ /(darwin)/i ? true : false ) unless defined?(Module.nesting[0]::OSX)
  WINE = ( ((!ENV["WINEPREFIX"].nil? || !ENV["WINELOADER"].nil? || File.exists?("Z:/usr/bin/wine")) && !OSX)? true : false ) unless defined?(Module.nesting[0]::WINE)
  WIN = ( !OSX && !WINE ) unless defined?(Module.nesting[0]::WIN)

  @error_OS = "UnityPreviews.rb is a plugin intended for Ubuntu/Linux users." <<
  "Unfortunately Wine could not be found. Your system has been detected as #{WIN ? 'Windows' : OSX ? 'OS X' : 'unknown'}"

class OpenObserver < Sketchup::AppObserver

  def onOpenModel(model)
    # Check if a thumbnail exists,
    UnityPreviews.get_thumb_path(model.path){|path|
      unless File.exists?(path)
        # if not, create a thumbnail.
        UnityPreviews.create(model)
      else
      end
    }
    # Observe when model is saved.
    model.add_observer(SaveObserver.new)
  end

  alias_method(:onNewModel, :onOpenModel)

end


class SaveObserver < Sketchup::ModelObserver

  def onPostSaveModel(model)
    # Create a thumbnail.
    t = UI.start_timer(0.5, false){ UI.stop_timer(t)
      UnityPreviews.create(model)
    }
  end

end


# Main method: Creates a thumbnail for the given model.
# @param [Sketchup::Model] model
#
def self.create(model)
  puts @error_OS unless WINE
  model_path = model.path
  # Disable ground and horizon for transparent export.
  r = model.rendering_options
  drawGround, r["DrawGround"] = r["DrawGround"], false
  drawHorizon, r["DrawHorizon"] = r["DrawHorizon"], false
  # Get the thumbnail path.
  self.get_thumb_path(model_path){|path|
    keys = {
      :filename => path,
      :width => 640, # TODO: preserve aspect ratio (?)
      :height => 480,
      :antialias => true,
      :compression => 0.9, # Actually only for jpeg, not used here.
      :transparent => true
    }
    model.active_view.write_image(keys)
    # Set thumbnail modification time to file modification time.
    File.utime(File.mtime(model_path), File.mtime(model_path), path) rescue nil
    # EDIT: Create also a normal thumbnail.
    path.sub!(/large/, "normal")
    keys[:filename] = path; keys[:width] = 128; keys[:height] = 128
    model.active_view.write_image(keys)
    File.utime(File.mtime(model_path), File.mtime(model_path), path) rescue nil
    # Reset ground and horizon.
    (r["DrawGround"] = drawGround; r["DrawHorizon"] = drawHorizon)
    true
  }
end


###########


# Returns the path of the thumbnail that corresponds to the given file.
# We need to convert the Windows/Wine file path into a native unix one and then turn it into a uri (file://).
# The thumbnail's filename will then be the md5 sum of the uri it is referring to.
# @param [String] model_path a file path
# @param [Proc] block a code block to execute with the thumbnail path as argument.
#
def self.get_thumb_path(model_path, &block)
  return false if model_path.empty? #|| !File.exists?(model_path)
  # platform: Wine/Linux
  # Convert Wine path into Linux path and resolve symbolic links.
  linux_path = %[$(readlink -f "$(winepath -u #{model_path.inspect})")]
  # Convert to URI. (better solution?)
  linux_uri = %[file://$(python -c "import urllib; print urllib.quote('''#{linux_path}''')")]
  # Create md5 of URI and attach to thumbnail directory.
  self.md5(linux_uri){|md5|
    thumbnail_dirs = [
      "Z:#{ENV['XDG_CACHE_HOME']}/thumbnails",
      "Z:/home/#{ENV['USER']}/.cache/thumbnails",
      "Z:/home/#{ENV['USER']}/.thumbnails"
    ]
    thumbnail_dir = thumbnail_dirs.find{|f| File.exists?(f)}
    break false unless thumbnail_dir
    large_dir = File.join(thumbnail_dir, "large")
    Dir.mkdir(large_dir) unless File.exists?(large_dir)
    path = File.join(large_dir, md5.to_s+".png")
    block.call(path)
  }
end


# (not needed)
# Encodes special characters in an URI.
# @param [String] string
# @return [String] the escaped string.
#
def self.escapeURI(string)
  unreserved="-_.!~*'()a-zA-Z\\d"
  reserved = ";/?:@&=+$,\\[\\]"
  unsafe = Regexp.new("[^#{unreserved}#{reserved}]", false, 'N')
  string.gsub(unsafe){|us|
    tmp = ''
    us.each_byte{|uc|
      tmp << sprintf('%%%02X', uc)
    }
    tmp
  }
end


# Get an md5 hash via asynchronous native system call.
# @platform Linux/Wine
# @note Ruby in Wine still freezes when using system()
def self.md5(data, &block)
  return false unless File.exists?("Z:/usr/bin/md5sum")
  cmd = "echo -n #{data} | md5sum"
  self.run_shell_command(cmd){|returnvalue|
    md5 = returnvalue[/^\S+/] # to strip of additional characters (normally file or " -")
    block.call(md5) unless md5.nil?
  }
end


# Execute a single command / a list of commands.
# @param [String] cmd command string or array containing command strings
# @param [Proc] code block to execute after the commands have finished
#
def self.run_shell_command(cmd, &block)
  cmd = [cmd] if cmd.class != Array
  # platform: Wine/Linux
  fsh = File.join(@temp, "temporary.sh")
  File.open(fsh, "w"){|f|
    f.puts %[ #!/bin/sh]
    f.print( cmd.join(";") )
    f.print %[ > "`wine winepath.exe -u '#{@temp}'`/result.txt";]
  }
  begin
    system(%[Z:\\bin\\bash "#{ENV["WINEPREFIX"]}/dosdevices/#{fsh.sub(/^(\w)(?=\:)/,@temp[0,1].downcase)}"])
  rescue
  end
  # since the batch script runs asynchronously, we need to know when it is finished
  # the batch script creates a "finished.txt" file, and we continue only when it is created
  self.file_observer(File.join(@temp, "result.txt")){
    File.open(File.join(@temp, "result.txt")){|file|
      result = file.read
      block.call(result)
    }
    File.delete(fsh) rescue nil
  }
end # def run_shell_command


# Observe when a file is created.
# @param [String] file file path
# @param [Proc] code block to execute when file is created
#
def self.file_observer(file, &block)
  if !File.exists?(file)
    UI.start_timer(0.2){ self.file_observer(file, &block) }
  else
    block.call
    File.delete(file) rescue nil
  end
end


if( not file_loaded?(__FILE__) )
  file_loaded(__FILE__)
  Sketchup.add_observer(OpenObserver.new)
  Sketchup.active_model.add_observer(SaveObserver.new)
end


end #module UnityPreviews



end #module AE
